//*******************************************************************************
//* FILE NAME: arm_control.c
//*
//* DESCRIPTION:
//*	Prototype control of the 2007 three joint turreted arm
//*******************************************************************************

#include <stdio.h>
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "user_routines.h" 
#include "arm_control.h"

//*******************************************************************************
//* FUNCTION NAME: arm_lookup
//* PURPOSE:       
//*	Lookup tables for proper relative arm joint positions
//*	As a time saver this is structured as a table indexed by a value of joint1 to the closest 10th.
//*	Corresponding joint2 values are then interpolated between the closest pair of joint1 ten's
//*	This just saves us from having to collect joint data for every possible position.
//* ARGUMENTS:
//*	position - HIGH, MEDIUM, LOW
//*	joint1 - where joint 1 currently is (or anticipated position to be based on OI command)
//*	joint2 - OUTPUT - The joint2 value corresponding to joint1 above
//*******************************************************************************
void arm_lookup(unsigned char position, unsigned int joint1, unsigned int *joint2)
{
	long temp_joint;

	// Interpolate between the two bounding values for all joints
	switch(position)
	{
		case HIGH:  // Tallest
			if (joint1 < JOINT1_HIGH_MIN)
			{
				*joint2 = JOINT2_HIGH_MIN; // Take the minimum value when out of range
			}
			else if (joint1 > JOINT1_HIGH_MAX)
			{
				*joint2 = JOINT2_HIGH_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint2 = (((long)(joint1-JOINT1_HIGH_MIN) * (JOINT2_HIGH_MAX - JOINT2_HIGH_MIN)) / (JOINT1_HIGH_MAX - JOINT1_HIGH_MIN))  + JOINT2_HIGH_MIN;  // Interpolate for joint2 value
			}
			break;

		case MEDIUM:
			if (joint1 < JOINT1_MEDIUM_MIN)
			{
				*joint2 = JOINT2_MEDIUM_MIN; // Take the minimum value when out of range
			}
			else if (joint1 > JOINT1_MEDIUM_MAX)
			{
				*joint2 = JOINT2_MEDIUM_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint2 = (((long)(joint1-JOINT1_MEDIUM_MIN) * (JOINT2_MEDIUM_MAX - JOINT2_MEDIUM_MIN)) / (JOINT1_MEDIUM_MAX - JOINT1_MEDIUM_MIN))  + JOINT2_MEDIUM_MIN;  // Interpolate for joint2 value
			}
			break;

		case LOW:
			if (joint1 < JOINT1_LOW_MIN)
			{
				*joint2 = JOINT2_LOW_MIN; // Take the minimum value when out of range
			}
			else if (joint1 > JOINT1_LOW_MAX)
			{
				*joint2 = JOINT2_LOW_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint2 = (((long)(joint1-JOINT1_LOW_MIN) * (JOINT2_LOW_MAX - JOINT2_LOW_MIN)) / (JOINT1_LOW_MAX - JOINT1_LOW_MIN))  + JOINT2_LOW_MIN;  // Interpolate for joint2 value
			}
			break;

		case FLOOR:	// Invader Zim  "so very short"
			if (joint1 < JOINT1_FLOOR_MIN)
			{
				*joint2 = JOINT2_FLOOR_MIN; // Take the minimum value when out of range
			}
			else if (joint1 > JOINT1_FLOOR_MAX)
			{
				*joint2 = JOINT2_FLOOR_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint2 =  (((long) (joint1-JOINT1_FLOOR_MIN) * (JOINT2_FLOOR_MAX - JOINT2_FLOOR_MIN)) / (JOINT1_FLOOR_MAX - JOINT1_FLOOR_MIN))  + JOINT2_FLOOR_MIN;  // Interpolate for joint2 value
			}
			break;
	}
 }

//*******************************************************************************
//* FUNCTION NAME: arm_lookup_2
//* PURPOSE:       
//*	The opposite of arm_lookup this one finds joint1 based on the joint2 position
//* ARGUMENTS:
//*	position - HIGH, MEDIUM, LOW
//*	joint2 - where joint 2 currently is (or anticipated position to be based on OI command)
//*	joint1 - OUTPUT - The joint1 value corresponding to join21 above
//*******************************************************************************
void arm_lookup_2(unsigned char position, unsigned int *joint1, unsigned int joint2)
{
	// Interpolate between the two bounding values for all joints
	switch(position)
	{
		case HIGH:  // Tallest
			if (joint2 < JOINT2_HIGH_MIN)
			{
				*joint1 = JOINT1_HIGH_MIN; // Take the minimum value when out of range
			}
			else if (joint2 > JOINT2_HIGH_MAX)
			{
				*joint1 = JOINT1_HIGH_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint1 = (((long)(joint2-JOINT2_HIGH_MIN) * (JOINT1_HIGH_MAX - JOINT1_HIGH_MIN)) / (JOINT2_HIGH_MAX - JOINT2_HIGH_MIN))  + JOINT1_HIGH_MIN;  // Interpolate for joint2 value
			}
			break;

		case MEDIUM:
			if (joint2 < JOINT2_MEDIUM_MIN)
			{
				*joint1 = JOINT1_MEDIUM_MIN; // Take the minimum value when out of range
			}
			else if (joint2 > JOINT2_MEDIUM_MAX)
			{
				*joint1 = JOINT1_MEDIUM_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint1 = (((long)(joint2-JOINT2_MEDIUM_MIN) * (JOINT1_MEDIUM_MAX - JOINT1_MEDIUM_MIN)) / (JOINT2_MEDIUM_MAX - JOINT2_MEDIUM_MIN))  + JOINT1_MEDIUM_MIN;  // Interpolate for joint2 value
			}
			break;

		case LOW:
			if (joint2 < JOINT2_LOW_MIN)
			{
				*joint1 = JOINT1_LOW_MIN; // Take the minimum value when out of range
			}
			else if (joint2 > JOINT2_LOW_MAX)
			{
				*joint1 = JOINT1_LOW_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint1 = (((long)(joint2-JOINT2_LOW_MIN) * (JOINT1_LOW_MAX - JOINT1_LOW_MIN)) / (JOINT2_LOW_MAX - JOINT2_LOW_MIN))  + JOINT1_LOW_MIN;  // Interpolate for joint2 value
			}
			break;

		case FLOOR:	// Invader Zim  "so very short"
			if (joint2 < JOINT2_FLOOR_MIN)
			{
				*joint1 = JOINT1_FLOOR_MIN; // Take the minimum value when out of range
			}
			else if (joint2 > JOINT2_FLOOR_MAX)
			{
				*joint1 = JOINT1_FLOOR_MAX; // Take the minimum value when out of range
			}
			else
			{
				*joint1 =  (((long) (joint2-JOINT2_FLOOR_MIN) * (JOINT1_FLOOR_MAX - JOINT1_FLOOR_MIN)) / (JOINT2_FLOOR_MAX - JOINT2_FLOOR_MIN))  + JOINT1_FLOOR_MIN;  // Interpolate for joint2 value
			}
			break;
	}
 }

//*******************************************************************************
//* FUNCTION NAME: arm_fold
//* PURPOSE:       
//*	Move the arm to it's home starting position
//*******************************************************************************
void arm_fold()
{
	// State Machine to do one (or more) joints in a particular order
	JOINT1_MOTOR = 127; JOINT2_MOTOR = 127;
}


//*******************************************************************************
//* FUNCTION NAME: arm_unfold
//* PURPOSE:       
//*	Move the arm to it's operational starting position
//*******************************************************************************
void arm_unfold()
{
	// State Machine to do one (or more) joints in a particular order
	JOINT1_MOTOR = 127; JOINT2_MOTOR = 127;
}


//*******************************************************************************
//* FUNCTION NAME: arm_pickup
//* PURPOSE:       
//*	Move the arm straight down to pickup a tube
//*******************************************************************************
void arm_dip()
{
	JOINT1_MOTOR = 127; JOINT2_MOTOR = 127;
}


//*******************************************************************************
//* FUNCTION NAME: arm_movement
//* PURPOSE:       
//*	Move the arm according to the requested joint positions
//*	Limit is 48" in short direction and 71" in the diagonal
//*******************************************************************************
#define JOINT_DEADBAND 2
#define MAX_JOINT1_SPEED 127  //was 40
#define MAX_JOINT2_SPEED 127

void arm_movement(unsigned int desired_joint1, unsigned int desired_joint2)
{
	unsigned int joint1, joint2;
	unsigned int kp1, kp2;
	long motor;
	unsigned int joint1_shouldbe;

	Coefficient_Input(&kp1, &kp2);

	//Current joint positions
	joint1 = Get_Analog_Value(rc_ana_in01);
	joint2 = Get_Analog_Value(rc_ana_in02);

	// ************* Joint 1
	if (joint1 < (desired_joint1 - JOINT_DEADBAND))
	{
		motor = (long)(desired_joint1-joint1)* kp1/100;
		if(motor > MAX_JOINT1_SPEED) 
		{
			motor = MAX_JOINT1_SPEED;
		}
		JOINT1_MOTOR = 127 + motor;
 	}
	else if(joint1 > (desired_joint1 + JOINT_DEADBAND))
	{
		motor = (long)(joint1-desired_joint1)* kp1/100;
		if(motor > MAX_JOINT1_SPEED) 
		{
			motor = MAX_JOINT1_SPEED;
		}
		JOINT1_MOTOR = 127 - motor;
	}
	else
	{
		JOINT1_MOTOR = 127;
	}

	// ************* Joint 2
	if (joint2 < (desired_joint2 - JOINT_DEADBAND))
	{
		motor = (long)(desired_joint2-joint2)* kp2/100;
		if(motor > MAX_JOINT2_SPEED) 
		{
			motor = MAX_JOINT2_SPEED;
		}
		motor = 127 + motor;
	}
	else if(joint2 > (desired_joint2 + JOINT_DEADBAND))
	{
		motor = (long)(joint2-desired_joint2)* kp2/100;
		if(motor > MAX_JOINT2_SPEED)
		{
			motor = MAX_JOINT2_SPEED;
		}
		motor = 127 - motor;
	}
	else
	{
		motor = 127;
	}
	JOINT2_MOTOR = (unsigned char) motor;

#ifdef MLM
	// ************* Slow/speed-up the joint(s) to match them
	// Judge all joints in reference to joint1
	// This will change based on which arm sections are dropping and which ones are rising against gravity
	// If joint1 is ahead of joint2 then slow joint1 based on the position error relative to joint2.
	// Avoid reversals and total stoppages.
	// Sort of reversed from the normal application of P, since if the error is small P should be small, but if the
	// error is large then P should be large.
	// So... determine who's ahead.

	arm_lookup_2(HIGH, &joint1_shouldbe, joint2);  // Where should joint1 be right now.

	// Is the joint leading or training where it should be?
	// Four cases: A) desired-shouldbe-current, B) desired-current-shouldbe, C) current-shouldbe-desired, D) shouldbe-current-desired
	// Ignore overshoots and let the system correct for that unimpeded
kp1 = 100;
//	motor = 0;
//	if ( (desired_joint1 > joint1_shouldbe) && (desired_joint1 > joint1) )
//	{ 
		motor =( (long)joint1_shouldbe - joint1) *  kp1/100;
//	}
//	else if ( (desired_joint1 < joint1_shouldbe) && (desired_joint1 < joint1) )
//	{
//		motor = joint1 - joint1_shouldbe *  kp1/100;
//	}
	if (JOINT1_MOTOR < (127 + motor) || JOINT1_MOTOR > (127 + motor))
		JOINT1_MOTOR += motor;
	printf("motor=%ld, joint1_shouldbe=%d\r", motor, joint1_shouldbe);
#endif

	// Safety override for limits of arm joints
	if (joint1 < JOINT1_MIN && JOINT1_MOTOR < 127) 
	{
		JOINT1_MOTOR = 127;
		printf("WARN ARM LIMIT\r");
	}
	else if ( joint1 > JOINT1_MAX && JOINT1_MOTOR > 127)
	{
		JOINT1_MOTOR = 127;
		printf("WARN ARM LIMIT\r");
	}
	if (joint2 < JOINT2_MIN && JOINT2_MOTOR < 127)
 	{
		JOINT1_MOTOR = 127;
		printf("WARN ARM LIMIT\r");
	}
	else if (joint2 > JOINT2_MAX && JOINT2_MOTOR > 127) 
	{
		JOINT1_MOTOR = 127;
		printf("WARN ARM LIMIT\r");
	}
}

void arm_fold(void);
void arm_dip(void);
void arm_unfold(void);

//*******************************************************************************
//* FUNCTION NAME: Arm_Control
//* PURPOSE:       Handles all arm joint movement
//* NOTES:   PWM_in1, etc.
//*******************************************************************************
void Arm_Control(void)
{
	unsigned int desired_joint1, desired_joint2;
	static char arm_position=HIGH;  // Arm starting position
	static char in5_flag = 1;

	static unsigned int counter=0;
	unsigned int kp1, kp2;
	unsigned int joint1, joint2, high, low;

	// pwm01 is the first joint
	// pwm02 is the second joint
	// rc_ana_in01 is the first joint potentiometer
	// rc_ana_in02 is the second joint's potentiometer

	// Drive the joints to particular positions according to the PWM_in1 joystick input
	// and user selected arm height

	// Select arm operating position
	// Only count a button push once, so wait for user to release it
	if (PWM_in5 > BUTTON_REV_THRESH && PWM_in5 < BUTTON_FWD_THRESH && in5_flag == 1)
	{
		in5_flag = 0;
		printf("arm_position=%d\r", arm_position);
	}
	else if (in5_flag == 0)
	{
		if (PWM_in5 < BUTTON_REV_THRESH)
		{
			in5_flag = 1;
			if (arm_position == HIGH) 
				arm_position = FLOOR;
			else
				arm_position--;
		}
		else if (PWM_in5 > BUTTON_FWD_THRESH)
		{
			in5_flag = 1;
			if (arm_position == FLOOR) 
				arm_position = HIGH;
			else
				arm_position++;
		}
	}

	joint1 = Get_Analog_Value(rc_ana_in01);
	joint2 = Get_Analog_Value(rc_ana_in02);

	switch (arm_position)
	{
		case FOLD:   // Arm in stowed position
			arm_fold();
			break;
		case UNFOLD:   // Arm unfold from starting position
			arm_unfold();
			break;
		case HIGH:   // Arm in high position
			// Scale OI input to match joint1 pot range
			desired_joint1 = (long)PWM_in2 * (JOINT1_HIGH_MAX - JOINT1_HIGH_MIN) / 255 + JOINT1_HIGH_MIN;
			arm_lookup(HIGH, desired_joint1, &desired_joint2);
			arm_lookup_2(HIGH, &desired_joint1, joint2);
			arm_movement(desired_joint1, desired_joint2); // Set the joint 1&2 motors
			break;
		case MEDIUM:   // Arm in medium position
			// Scale OI input to match joint1 pot range
			desired_joint1 = (long)PWM_in2 * (JOINT1_MEDIUM_MAX - JOINT1_MEDIUM_MIN) / 255 + JOINT1_MEDIUM_MIN;
			arm_lookup(MEDIUM, desired_joint1, &desired_joint2);
			arm_lookup_2(MEDIUM, &desired_joint1, joint2);
			arm_movement(desired_joint1, desired_joint2); // Set the joint 1&2 motors
			break;
		case LOW:   // Arm in low position
			// Scale OI input to match joint1 pot range
			desired_joint1 = (long)PWM_in2 * (JOINT1_LOW_MAX - JOINT1_LOW_MIN) / 255 + JOINT1_LOW_MIN;
			arm_lookup(LOW, desired_joint1, &desired_joint2);
			arm_lookup_2(LOW, &desired_joint1, joint2);
			arm_movement(desired_joint1, desired_joint2); // Set the joint 1&2 motors
			break;
		case FLOOR:   // Arm in floor position -hovering ready for tube pickup
			// Scale OI input to match joint1 pot range
			desired_joint1 = (long)PWM_in2 * (JOINT1_FLOOR_MAX - JOINT1_FLOOR_MIN) / 255 + JOINT1_FLOOR_MIN;
			arm_lookup(FLOOR, desired_joint1, &desired_joint2);
			arm_lookup_2(FLOOR, &desired_joint1, joint2);
			arm_movement(desired_joint1, desired_joint2); // Set the joint 1&2 motors
			break;
		case DIP:   // pickup from floor
			JOINT1_MOTOR = 127; JOINT2_MOTOR = 127;
			arm_dip();
			break;
		case TEST:   // Whatever needs looking at
			JOINT1_MOTOR = 127; JOINT2_MOTOR = 127;
			break;
	}

	counter++;
	if ((counter % 38) == 0)
	{
//		joint1 = Get_Analog_Value(rc_ana_in01);
//		joint2 = Get_Analog_Value(rc_ana_in02);
		printf("arm@%d ", arm_position);
		printf("PWM_in2=%d, ", PWM_in2);
		printf("j1m=%d, j2m=%d,", JOINT1_MOTOR, JOINT2_MOTOR);
	Coefficient_Input(&kp1, &kp2);
		printf("kp1=%d, kp2=%d, ", kp1, kp2);
		printf("j1, %d, j2, %d, ", joint1, joint2);
		printf("dj1=%d, dj2=%d\r", desired_joint1, desired_joint2);
//		printf("--User\r");
	}

} // End Arm_Control


//*******************************************************************************
//* FUNCTION NAME: Coefficient_Input
//* PURPOSE:       Inputs and scales pot inputs to use as changable coefficients
//* NOTES:   
//*******************************************************************************
void Coefficient_Input(unsigned int *kp1, unsigned int *kp2)
{
	// pot (0-255) maps directly into the kp range of (0-255)
	*kp1 = 254; // was 207
	*kp2 = 254;  // was 210

	// Get coefficients
	*kp1 = Get_Analog_Value(rc_ana_in03)>>2; // Drop the last couple of bits that will be oscillating wildly anyway.
	*kp2 = Get_Analog_Value(rc_ana_in04)>>2; // This gives us a value of 0-255
}
